-- $Id: OffersDB.lua 92964 2009-06-26 02:44:22Z augusto $
local io = io
local string = string
local os = os

local error = error

local lfs = require "lfs"
local oil = require "oil"
local orb = oil.orb

local FileStream = require "loop.serial.FileStream"

local Log = require "openbus.util.Log"

local oop = require "loop.base"

---
--Mecanismo de persistncia de ofertas de servio.
--
module("core.services.registry.OffersDB", oop.class)

FILE_SUFFIX = ".offer"
FILE_SEPARATOR = "/"

---
--Cria um banco de dados de ofertas de servio.
--
--@param databaseDirectory O diretrio que sera utilizado para armazenas as
--ofertas de servio.
--
--@return O banco de dados de ofertas de servio.
---
function __init(self, databaseDirectory)
  local mode = lfs.attributes(databaseDirectory, "mode")
  if not mode then
    Log:service("O diretorio ["..databaseDirectory.."] nao foi encontrado. "..
        "Criando...")
    local status, errorMessage = lfs.mkdir(databaseDirectory)
    if not status then
      Log:error("Nao foi possivel criar o diretorio ["..databaseDirectory.."]: "..errorMessage)
      error(errorMessage)
    end
  end
  return oop.rawnew(self, {
    databaseDirectory = databaseDirectory,
    dbOffers = {},
  })
end

---
--Obtm todas as ofertas de servio.
--
--@return As ofertas de servio.
---
function retrieveAll(self)
  local offerFiles = lfs.dir(self.databaseDirectory)
  local offerEntries = {}
  for fileName in offerFiles do
    if string.sub(fileName, -(#self.FILE_SUFFIX)) == self.FILE_SUFFIX then
      local offerFile = io.open(self.databaseDirectory..self.FILE_SEPARATOR..fileName)
      local stream = FileStream{
        file = offerFile
      }
      local offerEntry = stream:get()
      offerFile:close()

      self.dbOffers[offerEntry.identifier] = true

      -- caso especial para referencias a membros
      local memberIOR = offerEntry.offer.member
      offerEntry.offer.member = orb:newproxy(memberIOR)

      offerEntries[offerEntry.identifier] = offerEntry
    end
  end
  return offerEntries
end

---
--Insere uma entrada no banco de dados.
--
--@param offerEntry A oferta de servio.
--
--@return true caso a oferta tenha sido inserida, ou false e uma mensagem
--de erro, caso contrrio.
---
function insert(self, offerEntry)
  if self.dbOffers[offerEntry.identifier] then
    return false, "A oferta especificada ja existe."
  end
  local status, errorMessage = self:writeOffer(offerEntry)
  if not status then
    return false, errorMessage
  end
  self.dbOffers[offerEntry.identifier] = true
  return true
end

---
--Atualiza uma oferta de servio.
--
--@param offerEntry A oferta de servio.
--
--@return true caso a oferta seja atualizada, ou false e uma mensagem de erro,
--caso contrrio.
---
function update(self, offerEntry)
  if not self.dbOffers[offerEntry.identifier] then
    return false, "A oferta especificada no existe."
  end
  return self:writeOffer(offerEntry)
end

---
--Remove uma oferta de servio.
--
--@param offerEntry A oferta de servio.
--
--@return true caso a oferta tenha sido removida, ou false e uma mensagem de
--erro, caso contrrio.
---
function delete(self, offerEntry)
  if not self.dbOffers[offerEntry.identifier] then
    return false, "A oferta especificada no existe."
  end
  local status, errorMessage = self:removeOffer(offerEntry)
  if not status then
    return false, errorMessage
  end
  self.dbOffers[offerEntry.identifier] = nil
  return true
end

---
--Escreve a oferta de servio em arquivo.
--
--@param offerEntry A oferta de servio.
--
--@return true caso o arquivo seja gerado, ou false e uma mensagem de erro,
--caso contrrio.
---
function writeOffer(self, offerEntry)
  local offerFile, errorMessage =  io.open(self.databaseDirectory..
      self.FILE_SEPARATOR..offerEntry.identifier..self.FILE_SUFFIX, "w")
  if not offerFile then
    return false, errorMessage
  end
  local stream = FileStream{
    file = offerFile,
    getmetatable = false,
  }

  local member =  offerEntry.offer.member
  stream[member] = "'"..orb:tostring(member).."'"
  stream:put(offerEntry)
  offerFile:close()

  return true
end

---
--Remove o arquivo da oferta de servio.
--
--@param offerEntry A oferta de servio.
--
--@return Caso o arquivo no seja removido, retorna nil e uma mensagem de erro.
---
function removeOffer(self, offerEntry)
  return os.remove(self.databaseDirectory..self.FILE_SEPARATOR..
      offerEntry.identifier..self.FILE_SUFFIX)
end
