-- $Id: Session.lua 95899 2009-09-08 11:57:45Z rodrigoh $

local tostring = tostring
local ipairs = ipairs
local pairs = pairs

local oil = require "oil"
local orb = oil.orb

local luuid = require "uuid"

local Log = require "openbus.util.Log"

local oop = require "loop.base"

---
--Sesso compartilhada pelos membros associados a uma mesma credencial.
---
module "core.services.session.Session"

--------------------------------------------------------------------------------
-- Faceta ISession
--------------------------------------------------------------------------------

Session = oop.class{sessionMembers = {}}

---
--Obt�m o identificador da sesso.
--
--@return O identificador da sesso.
---
function Session:getIdentifier()
  return self.identifier
end

---
--Adiciona um membro a sesso.
--
--@param member O membro a ser adicionado.
--
--@return O identificador do membro na sesso.
---
function Session:addMember(member)
  local memberName = member:getComponentId().name
  Log:service("Membro "..memberName.." adicionado a sesso")
  local identifier = self:generateMemberIdentifier()
  self.sessionMembers[identifier] = member

  -- verifica se o membro recebe eventos
  local eventSinkInterface = "IDL:openbusidl/ss/SessionEventSink:1.0"
  local eventSink = member:getFacet(eventSinkInterface)
  if eventSink then
    Log:service("Membro "..memberName.." receber eventos")
    local eventSinks = self.context.SessionEventSink.eventSinks
    eventSinks[identifier] =  orb:narrow(eventSink,
        eventSinkInterface)
  else
    Log:service("Membro "..memberName.." no receber eventos")
  end
  return identifier
end

---
--Remove um membro da sesso.
--
--@param identifier O identificador do membro na sesso.
--
--@return true caso o membro tenha sido removido da sesso, ou false caso
--contrrio.
---
function Session:removeMember(identifier)
  member = self.sessionMembers[identifier]
  if not member then
    return false
  end
  Log:service("Membro "..member:getComponentId().name.." removido da sesso")
  self.sessionMembers[identifier] = nil
  self.context.SessionEventSink.eventSinks[identifier] = nil
  return true
end

---
--Obtm a lista de membros de uma sesso.
--
--@return Os membros da sesso.
---
function Session:getMembers()
  local members = {}
  for _, member in pairs(self.sessionMembers) do
    table.insert(members, member)
  end
  return members
end

---
--Gera um identificador de membros de sesso.
--
--@return O identificador de membro de sesso.
---
function Session:generateMemberIdentifier()
  return luuid.new("time")
end

--------------------------------------------------------------------------------
-- Faceta SessionEventSink
--------------------------------------------------------------------------------

SessionEventSink = oop.class{eventSinks = {}}

---
--Repassa evento para membros da sesso.
--
--@param event O evento.
---
function SessionEventSink:push(event)
  Log:service("Repassando evento "..event.type.." para membros de sesso")
  for _, sink in pairs(self.eventSinks) do
    local result, errorMsg = oil.pcall(sink.push, sink, event)
    if not result then
      Log:service("Erro ao enviar evento para membro de sesso: "..errorMsg)
    end
  end
end

---
--Solicita a desconexo de todos os membros da sesso.
---
function SessionEventSink:disconnect()
  Log:service("Desconectando os membros da sesso")
  for _, sink in pairs(self.eventSinks) do
    local result, errorMsg = oil.pcall(sink.disconnect, sink)
    if not result then
      Log:service("Erro ao tentar desconectar membro de sesso: "..errorMsg)
    end
  end
end


