-- $Id: Openbus.lua 90798 2009-04-23 02:06:47Z augusto $

local oil = require "oil"
local oop = require "loop.base"
local lce = require "lce"
local log = require "openbus.util.Log"
local CredentialManager = require "openbus.util.CredentialManager"
local Utils = require "openbus.util.Utils"
local LeaseRenewer = require "openbus.lease.LeaseRenewer"

local pairs = pairs
local os = os
local loadfile = loadfile
local assert = assert
local require = require

---
-- API de acesso a um barramento OpenBus.
---
module "openbus.Openbus"

Openbus = oop.class {
  ---
  -- A instncia nica do barramento.
  ---
  _instance = nil,
  ---
  -- O ORB.
  ---
  orb = nil,
  ---
  -- Indica se o ORB j foi finalizado.
  ---
  isORBFinished = true,
  ---
  -- O host do Servio de Controle de Acesso.
  ---
  host = nil,
  ---
  -- A porta do Servio de Controle de Acesso.
  ---
  port = -1,
  ---
  -- O Servio de Controle de Acesso.
  ---
  acs = nil,
  ---
  -- Interface ILeaseProvider do Servio de Controle de Acesso.
  ---
  lp = nil,
  ---
  -- O renovador do <i>lease</i>.
  ---
  leaseRenewer = nil,
  ---
  -- <i>Callback</i> para a notificao de que um <i>lease</i> expirou.
  ---
  leaseExpiredCallback = nil,
  ---
  -- Servio de registro.
  ---
  rgs = nil,
  ---
  -- Servio de sesso.
  ---
  ss = nil,
  ---
  -- Interceptador servidor.
  ---
  serverInterceptor = nil,
  ---
  -- Configurao do interceptador servidor.
  ---
  serverInterceptorConfig = nil,
  ---
  -- Interceptador cliente.
  ---
  clientInterceptor = nil,
  ---
  -- Configurao do interceptador cliente.
  ---
  clientInterceptorConfig = nil,
  ---
  -- Credencial recebida ao se conectar ao barramento.
  ---
  credentialManager = nil,
  ---
  -- Indica o estado da conexo. 1 = conectado, 2 = desconectado
  ---
  connectionState = 2,
}

---
-- Fornece a instncia nica do barramento.
--
--@return A instncia nica do barramento.
---
function Openbus:__init()
  if not self._instance then
    self._instance = oop.rawnew(self, {})
  end
  return self._instance
end

---
-- Retorna ao seu estado inicial, ou seja, desfaz as definies de atributos
-- realizadas.
---
function Openbus:_resetAll()
  self.credentialManager = nil
  if not self.isORBFinished and self.orb then
    self:finish()
  end
  self.orb = nil
  self.isORBFinished = false
  self.acs = nil
  self.host = nil
  self.port = -1
  self.lp = nil
  self.leaseRenewer = nil
  self.leaseExpiredCallback = nil
  self.rgs = nil
  self.ss = nil
  self.serverInterceptor = nil
  self.serverInterceptorConfig = nil
  self.clientInterceptor = nil
  self.clientInterceptorConfig = nil
  self.connectionState = 2
end

---
-- Retorna ao seu estado inicial, ou seja, desfaz as definies de atributos
-- realizadas.
---
function Openbus:_reset()
  self.acs = nil
  self.rgs = nil
  self.lp = nil
  self.ss = nil
  self.leaseRenewer = nil
  self.connectionState = 2
  if self.credentialManager then
    self.credentialManager.invalidate()
    self.credentialManager.invalidateThreadValue()
  end
end

---
-- Cadastra os interceptadores cliente e servidor no ORB.
---
function Openbus:_setInterceptors()
  self.credentialManager = CredentialManager()
  local config
  if not self.serverInterceptorConfig or not self.clientInterceptorConfig then
    local DATA_DIR = os.getenv("OPENBUS_DATADIR")
    config = assert(loadfile(DATA_DIR ..
      "/conf/advanced/InterceptorsConfiguration.lua"))()
  end
  self.serverInterceptor = ServerInterceptor(self.serverInterceptorConfig or
    config, self.acs)
  self.orb:setserverinterceptor(self.serverInterceptor)
  self.clientInterceptor = ClientInterceptor(self.clientInterceptorConfig or
    config, self.credentialManager)
  self.orb:setclientinterceptor(self.clientInterceptor)
end

---
-- Obtm as facetas do Servio de Controle de Acesso. Este
-- mtodo tambm instancia um observador de <i>leases</i>.
--
-- @return {@code false} caso ocorra algum erro, {@code true} caso contrrio.
---
function Openbus:_fetchACS()
  local status, acs, lp = oil.pcall(Utils.fetchAccessControlService, self.orb,
    self.host, self.port)
  if not status then
    log:error("Erro ao obter as facetas do Servio de Controle de Acesso." ..
      "Erro: " .. acs)
    return false
  end
  if not acs then
    -- erro j foi pego e logado pela Utils
    return false
  end
  self.acs, self.lp = acs, lp
  if not self.clientInterceptor or not self.serverInterceptor then
    local status, err = oil.pcall(self._setInterceptors, self)
    if not status then
      log:error("Erro ao cadastrar interceptadores no ORB. Erro: " .. err)
      return false
    end
  end
  return true
end

---
-- Carrega as IDLs dos servios bsicos e do SCS.
---
function Openbus:_loadIDLs()
  local IDLPATH_DIR = os.getenv("IDLPATH_DIR")
  if not IDLPATH_DIR then
    log:error("Openbus: A varivel IDLPATH_DIR no foi definida.")
    return false
  end
  local idlfile = IDLPATH_DIR .. "/scs.idl"
  self.orb:loadidlfile(idlfile)
  idlfile = IDLPATH_DIR .. "/access_control_service.idl"
  self.orb:loadidlfile(idlfile)
  idlfile = IDLPATH_DIR .. "/registry_service.idl"
  self.orb:loadidlfile(idlfile)
  idlfile = IDLPATH_DIR .. "/session_service.idl"
  self.orb:loadidlfile(idlfile)
--  idlfile = IDLPATH_DIR .. "/data_service.idl"
--  self.orb:loadidlfile(idlfile)
  return true
end

---
-- Finaliza o procedimento de conexo, aps um login bem sucedido. Salva a
-- credencial e inicia o processo de renovao de lease.
--
-- @param credential A credencial do membro.
-- @param lease O perodo de tempo entre as renovaes do lease.
--
-- @return O servio de registro. {@code false} caso ocorra algum erro.
---
function Openbus:_completeConnection(credential, lease)
  self.credentialManager:setValue(credential)
  self.leaseRenewer = LeaseRenewer(lease, credential, self.lp, self)
  self.leaseRenewer:startRenew()
  self.connectionState = 1
  return self:getRegistryService()
end

---
-- Retorna o barramento para o seu estado inicial, ou seja, desfaz as
-- definies de atributos realizadas. Em seguida, inicializa o Orb.
--
-- @param host Endereo do Servio de Controle de Acesso.
-- @param port Porta do Servio de Controle de Acesso.
-- @param props Conjunto de propriedades para a criao do ORB.
-- @param serverInterceptorConfig Configurao opcional do interceptador
--        servidor.
-- @param clientInterceptorConfig Configurao opcional do interceptador
--        cliente.
--
-- @return {@code false} caso ocorra algum erro, {@code true} caso contrrio.
---
function Openbus:resetAndInitialize(host, port, props, serverInterceptorConfig,
  clientInterceptorConfig)
  if not host then
    log:error("OpenBus: O campo 'host' no pode ser nil")
    return false
  end
  if not port or port < 0 then
    log:error("OpenBus: O campo 'port' no pode ser nil nem negativo.")
    return false
  end
  self:_resetAll()
  -- init
  self.host = host
  self.port = port
  self.serverInterceptorConfig = serverInterceptorConfig
  self.clientInterceptorConfig = clientInterceptorConfig
  -- configurao do OiL
  if not props then
    props = {}
  end
  if not props.tcpoptions then
    props.tcpoptions = {}
  end
  if not props.tcpoptions.reuseaddr then
    props.tcpoptions.reuseaddr = true
  end
  if not props.flavor then
    props.flavor = "intercepted;corba;typed;cooperative;base"
  end
  self.orb = oil.init(props)
  --TODO: remover esse uso de oil.orb no Openbus e mover os requires abaixo para
  --      o topo.
  oil.orb = self.orb
  ClientInterceptor = require "openbus.interceptors.ClientInterceptor"
  ServerInterceptor = require "openbus.interceptors.ServerInterceptor"
  -- carrega IDLs
  local status, err = oil.pcall(self._loadIDLs, self)
  if not status then
    log:error("OpenBus: Erro ao carregar as IDLs. Erro: " .. err)
    return false
  end
  return err
end

---
-- Fornece o ORB.
--
-- @return O ORB.
---
function Openbus:getORB()
  return self.orb
end

---
-- Executa o ORB.
---
function Openbus:run()
  oil.newthread(self.orb.run, self.orb)
end

---
-- Finaliza a execuo do ORB.
---
function Openbus:finish()
  if not self.orb then
    return
  end
  self.orb:shutdown()
  self.isORBFinished = true
end

---
-- Fornece o Servio de Controle de Acesso.
--
-- @return O Servio de Controle de Acesso.
---
function Openbus:getAccessControlService()
  return self.acs
end

---
-- Fornece o Servio de Registro. Caso o Openbus ainda no tenha a referncia
-- para o Servio de Registro, obtm a mesma a partir do Servio de Controle de
-- Acesso.
--
-- @return O Servio de Registro. Nil caso ainda no tenha sido obtida uma
--         referncia e o Servio de Controle de Acesso estiver inacessvel.
---
function Openbus:getRegistryService()
  if not self.rgs and self.acs then
    local status, err = oil.pcall(self.acs.getRegistryService, self.acs)
    if not status then
      log:warn("OpenBus: No foi possvel obter o Servio de Registro. Erro: " ..
        err)
      return nil
    end
    self.rgs = err
  end
  return self.rgs
end

---
-- Fornece o Servio de Sesso. Caso o Openbus ainda no tenha a referncia
-- para o Servio de Sesso, obtm a mesma a partir do Servio de Registro.
--
-- @return O Servio de Sesso. Nil caso ainda no tenha sido obtida uma
--         referncia e o Servio de Controle de Acesso estiver inacessvel.
---
function Openbus:getSessionService()
  if not self.rgs then
    self:getRegistryService()
  end
  if not self.ss and self.rgs then
    local facets = { Utils.SESSION_SERVICE_FACET_NAME }
    local offers = self.rgs.find(facets)
    if not offers or #offers > 0 then
      local component = self.orb:narrow(offers[1].member)
      local facet = component:getFacet(Utils.SESSION_SERVICE_INTERFACE)
      if not facet then
        return nil
      end
      self.ss = self.orb:narrow(facet)
      return self.ss
    end
    return nil
  end
  return self.ss
end

---
-- Fornece a credencial da entidade.
--
-- @return A credencial.
---
function Openbus:getCredential()
  if not self.credentialManager then
    return nil
  end
  return self.credentialManager:getValue()
end

---
-- Define uma credencial a ser utilizada no lugar da credencial corrente. til
-- para fornecer uma credencial com o campo delegate preenchido.
--
-- @param credential Credencial a ser usada nas requisies a serem
--        realizadas.
---
function Openbus:setThreadCredential(credential)
  self.credentialManager:setThreadValue(credential)
end

---
-- Fornece a credencial interceptada a partir da requisio atual.
--
-- @return A credencial da requisio.
---
function Openbus:getInterceptedCredential()
  if not self.serverInterceptor then
    return nil
  end
  return self.serverInterceptor:getCredential()
end

---
-- Realiza uma tentativa de conexo com o barramento (servio de controle de
-- acesso e o servio de registro), via nome de usurio e senha.
--
-- @param user Nome do usurio.
-- @param password Senha do usurio.
--
-- @return O servio de registro. {@code false} caso ocorra algum erro.
--
-- @throws InvalidCredentialException Caso a credencial seja rejeitada ao
--         tentar obter o Servio de Registro.
---
function Openbus:connect(user, password)
  if not user or not password then
    log:error("OpenBus: Os parmetros 'user' e 'password' no podem ser nil.")
    return false
  end
  if self.connectionState == 2 then
    if not self.acs then
      if not self:_fetchACS() then
        log:error("OpenBus: No foi possvel acessar o barramento.")
        return false
      end
    end
    local success, credential, lease = self.acs:loginByPassword(user, password)
    if success then
      return self:_completeConnection(credential, lease)
    else
      log:error("OpenBus: No foi possvel conectar ao barramento.")
      return false
    end
  else
    log:error("OpenBus: O barramento j est conectado.")
    return false
  end
end

---
-- Realiza uma tentativa de conexo com o barramento (servio de controle de
-- acesso e o servio de registro), via certificado.
--
-- @param name Nome do usurio.
-- @param privateKey Chave privada.
-- @param acsCertificate Certificado a ser fornecido ao Servio de Controle de
--        Acesso.
--
-- @return O Servio de Registro. {@code false} caso ocorra algum erro.
--
-- @throws InvalidCredentialException Caso a credencial seja rejeitada ao
--         tentar obter o Servio de Registro.
---
function Openbus:connectByCertificate(name, privateKeyFile, acsCertificateFile)
  if not name or not privateKeyFile or not acsCertificateFile then
    log:error("OpenBus: Nenhum parmetro pode ser nil.")
    return false
  end
  if self.connectionState == 2 then
    if not self.acs then
      if not self:_fetchACS() then
        log:error("OpenBus: No foi possvel acessar o barramento.")
        return false
      end
    end
    local challenge = self.acs:getChallenge(name)
    if not challenge or #challenge == 0 then
      log:error("OpenBus: o desafio para " .. name ..
        " no foi obtido junto ao Servico de Controle de Acesso.")
      return false
    end
    local privateKey, err = lce.key.readprivatefrompemfile(privateKeyFile)
    if not privateKey then
      log:error("OpenBus: erro obtendo chave privada para ".. name .. " em " ..
        privateKeyFile .. ": " .. err)
      return false
    end
    local answer = lce.cipher.decrypt(privateKey, challenge)
    local certificate = lce.x509.readfromderfile(acsCertificateFile)
    answer = lce.cipher.encrypt(certificate:getpublickey(), answer)
    local success, credential, lease = self.acs:loginByCertificate(name, answer)
    if success then
      return self:_completeConnection(credential, lease)
    else
      log:error("OpenBus: Falha no login de " .. name)
      return false
    end
  else
    log:error("OpenBus: O barramento j est conectado.")
    return false
  end
end

---
-- Realiza uma tentativa de conexo com o barramento(servio de controle de
-- acesso e o servio de registro), a partir de uma credencial.
--
-- @param credential A credencial.
--
-- @return O servio de registro. {@code false} caso ocorra algum erro.
--
-- @throws ACSUnavailableException Caso o Servio de Controle de Acesso no
--         consiga ser contactado.
---
function Openbus:connectByCredential(credential)
  if not credential then
    log:error("OpenBus: O parmetro 'credential' no pode ser nil.")
    return false
  end
  if not self.acs then
    if not self:_fetchACS() then
      log:error("OpenBus: No foi possvel acessar o barramento.")
      return false
    end
  end
  if self.acs:isValid(credential) then
    self.credentialManager:setValue(credential)
    return self:getRegistryService()
  end
  log:error("OpenBus: Credencial invlida.")
  return false
end

---
-- Desfaz a conexo atual.
--
-- @return {@code true} caso a conexo seja desfeita, ou {@code false} se
--         nenhuma conexo estiver ativa ou ocorra um erro.
---
function Openbus:disconnect()
  if self.connectionState == 1 then
    if self.leaseRenewer then
      local status, err = oil.pcall(self.leaseRenewer.stopRenew,
        self.leaseRenewer)
      if not status then
        self.connectionState = 1
        log:error(
          "OpenBus: No foi possvel parar a renovao de lease. Erro: " .. err)
        return false
      end
      self.leaseRenewer = nil
    end
    status, err = oil.pcall(self.acs.logout, self.acs,
      self.credentialManager:getValue())
    if not status then
      self.connectionState = 1
      log:error("OpenBus: No foi possvel realizar o logout. Erro " .. err)
      return false
    else
      self:_resetAll()
    end
    return true
  else
    return false
  end
end

---
-- Informa o estado de conexo com o barramento.
--
-- @return {@code true} caso a conexo esteja ativa, ou {@code false}, caso
--         contrrio.
---
function Openbus:isConnected()
  return (self.connectionState == 1)
end

---
-- Adiciona um observador para receber eventos de expirao do <i>lease</i>.
--
-- @param lec O observador.
--
---
function Openbus:addLeaseExpiredCallback(lec)
  self.leaseExpiredCallback = lec
end

---
-- Remove um observador de expirao do <i>lease</i>.
--
---
function Openbus:removeLeaseExpiredCallback()
  self.leaseExpiredCallback = nil
end

---
-- Mtodo interno da API que recebe a notificao de que o lease expirou.
-- Deve deixar a classe em um estado em que a callback do usurio
-- possa se reconectar.
---
function Openbus:expired()
  -- Deve dar reset antes de chamar o usurio
  self:_reset()
  if self.leaseExpiredCallback then
    self.leaseExpiredCallback:expired()
  end
end

return Openbus()
